# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""It contains everything about certification via a file-based provider."""

import threading
from datetime import datetime
from typing import Optional
from xml.etree import ElementTree

from tos.credential import Credentials, CredentialsProvider

from tosfs.core import logger
from tosfs.exceptions import TosfsCertificationError

CERTIFICATION_REFRESH_INTERVAL_MINUTES = 60
CERTIFICATION_MAX_VALID_PERIOD_HOURS = 12


class FileCredentialsProvider(CredentialsProvider):
    """The class provides the credentials from a file.

    The file should be in the format of:
      <configuration>
        <property>
            <name>fs.tos.access-key-id</name>
            <value>access_key</value>
        </property>
        <property>
            <name>fs.tos.secret-access-key</name>
            <value>secret_key</value>
        </property>
        <property>
            <name>fs.tos.session-token</name>
            <value>session_token</value>
        </property>
      </configuration>

    It can only receive a file path which exists in the local file system.

    Note :
    1. The default refresh interval is 60 minutes.
    2. The maximum valid period for provisional certification is 12 hours.

    This provider will cache the credentials and refresh them every 60 minutes.
    And note that, it only reads the credentials from the file and refreshes itself,
    to guarantee the credentials are always up-to-date,
    we need a service update it internally.

    Examples
    --------
    >>> from tosfs import TosFileSystem
    >>> tosfs = TosFileSystem(
    >>>             endpoint="tos-cn-beijing.volcengine.com",
    >>>             regions="cn-bejing",
    >>>             credentials_provider=FileCredentialsProvider("dummy_path"))
    >>> tosfs.ls("tos://bucket/path")

    """

    def __init__(
        self,
        file_path: str,
        refresh_interval_min: int = CERTIFICATION_REFRESH_INTERVAL_MINUTES,
    ) -> None:
        """Initialize the FileCredentialsProvider."""
        self.file_path = file_path
        self.refresh_interval = refresh_interval_min
        if self.refresh_interval <= 0:
            logger.warning(
                f"Invalid refresh interval {self.refresh_interval}, "
                f"set to default value: 60 minutes"
            )
            self.refresh_interval = CERTIFICATION_REFRESH_INTERVAL_MINUTES
        if self.refresh_interval > CERTIFICATION_MAX_VALID_PERIOD_HOURS * 60:
            logger.warning(
                f"Invalid refresh interval {self.refresh_interval}, "
                f"set to maximum value: 60 minutes"
            )
            self.refresh_interval = CERTIFICATION_REFRESH_INTERVAL_MINUTES
        self.prev_refresh_time: Optional[datetime] = None
        self.credentials = None
        self._lock = threading.Lock()

    def get_credentials(self) -> Credentials:
        """Get the credentials from the file.

        Returns
        -------
        Credentials: The credentials object.

        Raises
        ------
        TosfsCertificationError: If the credentials cannot be retrieved.

        """
        res = self._try_get_credentials()
        if res is not None:
            return res
        with self._lock:
            try:
                res = self._try_get_credentials()
                if res is not None:
                    return res

                with open(self.file_path, "r") as f:
                    logger.debug(
                        f"Trying to refresh the credentials from file: "
                        f"{self.file_path}"
                    )
                    tree = ElementTree.parse(f)  # noqa S314
                    root = tree.getroot()

                    access_key_element = root.find(
                        ".//property[name='fs.tos.access-key-id']/value"
                    )
                    secret_key_element = root.find(
                        ".//property[name='fs.tos.secret-access-key']/value"
                    )
                    session_token_element = root.find(
                        ".//property[name='fs.tos.session-token']/value"
                    )

                    access_key = (
                        access_key_element.text
                        if access_key_element is not None
                        else None
                    )
                    secret_key = (
                        secret_key_element.text
                        if secret_key_element is not None
                        else None
                    )
                    session_token = (
                        session_token_element.text
                        if session_token_element is not None
                        else None
                    )

                    if None in (
                        access_key,
                        secret_key,
                        session_token,
                    ):
                        raise TosfsCertificationError(
                            "Missing required credential elements in the file"
                        )

                    self.prev_refresh_time = datetime.now()
                    self.credentials = Credentials(
                        access_key, secret_key, session_token
                    )
                    logger.debug(
                        f"Successfully refreshed the credentials from file: "
                        f"{self.file_path}"
                    )

                return self.credentials
            except Exception as e:
                raise TosfsCertificationError("Get certification error: ") from e

    def _try_get_credentials(self) -> Optional[Credentials]:
        if self.prev_refresh_time is None or self.credentials is None:
            return None
        if (
            datetime.now() - self.prev_refresh_time
        ).total_seconds() / 60 > CERTIFICATION_REFRESH_INTERVAL_MINUTES:
            logger.debug(
                f"Credentials are expired, "
                f"will try to refresh the credentials from file: "
                f"{self.file_path}"
            )
            return None
        return self.credentials



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from time import sleep

from tosfs.utils import random_str


def test_write_breakpoint_continuation(tosfs, bucket, temporary_workspace):
    file_name = f"{random_str()}.txt"
    first_part = random_str(10 * 1024 * 1024)
    second_part = random_str(10 * 1024 * 1024)

    with tosfs.open(f"{bucket}/{temporary_workspace}/{file_name}", "w") as f:
        f.write(first_part)
        # mock a very long block(business processing or network issue)
        sleep(60 * 15)
        f.write(second_part)

    assert tosfs.info(f"{bucket}/{temporary_workspace}/{file_name}")["size"] == len(
        first_part + second_part
    )

    with tosfs.open(f"{bucket}/{temporary_workspace}/{file_name}", "r") as f:
        assert f.read() == first_part + second_part


def test_read_breakpoint_continuation(tosfs, bucket, temporary_workspace):
    file_name = f"{random_str()}.txt"
    first_part = random_str(10 * 1024 * 1024)
    second_part = random_str(10 * 1024 * 1024)

    with tosfs.open(f"{bucket}/{temporary_workspace}/{file_name}", "w") as f:
        f.write(first_part)
        f.write(second_part)

    with tosfs.open(f"{bucket}/{temporary_workspace}/{file_name}", "r") as f:
        read_first_part = f.read(10 * 1024 * 1024)
        assert read_first_part == first_part
        # mock a very long block(business processing or network issue)
        sleep(60 * 15)
        read_second_part = f.read(10 * 1024 * 1024)
        assert read_second_part == second_part
        assert read_first_part + read_second_part == first_part + second_part



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The module contains the MultipartUploader class for the tosfs package."""

import io
import itertools
import os
import tempfile
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, Optional

from tos.models2 import CreateMultipartUploadOutput, PartInfo

from tosfs.retry import retryable_func_executor

if TYPE_CHECKING:
    from tosfs.core import TosFileSystem


class StagingPartMgr:
    """A class to handle staging parts for multipart upload."""

    def __init__(self, part_size: int, staging_dirs: itertools.cycle):
        """Instantiate a StagingPart object."""
        self.part_size = part_size
        self.staging_dirs = staging_dirs
        self.staging_buffer = io.BytesIO()
        self.staging_files: list[str] = []

    def write_to_buffer(self, chunk: bytes) -> None:
        """Write data to the staging buffer."""
        self.staging_buffer.write(chunk)
        if self.staging_buffer.tell() >= self.part_size:
            self.flush_buffer(False)

    def flush_buffer(self, final: bool = False) -> None:
        """Flush the staging buffer."""
        if self.staging_buffer.tell() == 0:
            return

        buffer_size = self.staging_buffer.tell()
        self.staging_buffer.seek(0)

        while buffer_size >= self.part_size:
            staging_dir = next(self.staging_dirs)
            with tempfile.NamedTemporaryFile(delete=False, dir=staging_dir) as tmp:
                tmp.write(self.staging_buffer.read(self.part_size))
                self.staging_files.append(tmp.name)
                buffer_size -= self.part_size

        if not final:
            remaining_data = self.staging_buffer.read()
            self.staging_buffer = io.BytesIO()
            self.staging_buffer.write(remaining_data)
        else:
            staging_dir = next(self.staging_dirs)
            with tempfile.NamedTemporaryFile(delete=False, dir=staging_dir) as tmp:
                tmp.write(self.staging_buffer.read())
                self.staging_files.append(tmp.name)
            self.staging_buffer = io.BytesIO()

    def get_staging_files(self) -> list[str]:
        """Get the staging files."""
        return self.staging_files

    def clear_staging_files(self) -> None:
        """Clear the staging files."""
        self.staging_files = []


class MultipartUploader:
    """A class to upload large files to the object store using multipart upload."""

    def __init__(
        self,
        fs: "TosFileSystem",
        bucket: str,
        key: str,
        part_size: int,
        thread_pool_size: int,
        multipart_threshold: int,
    ):
        """Instantiate a MultipartUploader object."""
        self.fs = fs
        self.bucket = bucket
        self.key = key
        self.part_size = part_size
        self.thread_pool_size = thread_pool_size
        self.multipart_threshold = multipart_threshold
        self.executor = ThreadPoolExecutor(max_workers=self.thread_pool_size)
        self.staging_part_mgr = StagingPartMgr(
            part_size, itertools.cycle(fs.multipart_staging_dirs)
        )
        self.parts: list = []
        self.mpu: CreateMultipartUploadOutput = None

    def initiate_upload(self) -> None:
        """Initiate the multipart upload."""
        self.mpu = retryable_func_executor(
            lambda: self.fs.tos_client.create_multipart_upload(self.bucket, self.key),
            max_retry_num=self.fs.max_retry_num,
        )

    def upload_multiple_chunks(self, buffer: Optional[io.BytesIO]) -> None:
        """Upload multiple chunks of data to the object store."""
        if buffer:
            buffer.seek(0)
            while True:
                chunk = buffer.read(self.part_size)
                if not chunk:
                    break
                self.staging_part_mgr.write_to_buffer(chunk)

    def upload_staged_files(self) -> None:
        """Upload the staged files to the object store."""
        self.staging_part_mgr.flush_buffer(True)
        futures = []
        for i, staging_file in enumerate(self.staging_part_mgr.get_staging_files()):
            part_number = i + 1
            futures.append(
                self.executor.submit(
                    self._upload_part_from_file, staging_file, part_number
                )
            )

        for future in futures:
            part_info = future.result()
            self.parts.append(part_info)

        self.staging_part_mgr.clear_staging_files()

    def _upload_part_from_file(self, staging_file: str, part_number: int) -> PartInfo:
        with open(staging_file, "rb") as f:
            content = f.read()

        out = retryable_func_executor(
            lambda: self.fs.tos_client.upload_part(
                bucket=self.bucket,
                key=self.key,
                part_number=part_number,
                upload_id=self.mpu.upload_id,
                content=content,
            ),
            max_retry_num=self.fs.max_retry_num,
        )

        os.remove(staging_file)
        return PartInfo(
            part_number=part_number,
            etag=out.etag,
            part_size=len(content),
            offset=None,
            hash_crc64_ecma=None,
            is_completed=None,
        )

    def complete_upload(self) -> None:
        """Complete the multipart upload."""
        retryable_func_executor(
            lambda: self.fs.tos_client.complete_multipart_upload(
                self.bucket,
                self.key,
                upload_id=self.mpu.upload_id,
                parts=self.parts,
            ),
            max_retry_num=self.fs.max_retry_num,
        )

    def abort_upload(self) -> None:
        """Abort the multipart upload."""
        if self.mpu:
            retryable_func_executor(
                lambda: self.fs.tos_client.abort_multipart_upload(
                    self.bucket, self.key, self.mpu.upload_id
                ),
                max_retry_num=self.fs.max_retry_num,
            )
            self.mpu = None



# ByteDance Volcengine EMR, Copyright 2024.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""The compatible module about AbstractFileSystem in fsspec."""
import os
import re
from concurrent.futures import ThreadPoolExecutor, as_completed
from typing import Any, Optional, Union

from fsspec import AbstractFileSystem
from fsspec.implementations.local import LocalFileSystem
from fsspec.utils import other_paths

magic_check_bytes = re.compile(b"([*?[])")
magic_check = re.compile("([*?[])")


def has_magic(s: str) -> bool:
    """Check if a string has glob characters."""
    if isinstance(s, bytes):
        match = magic_check_bytes.search(s)
    else:
        match = magic_check.search(s)
    return match is not None


class EnhancedLocalFileSystem(LocalFileSystem):
    """Enhanced LocalFileSystem than fsspec's LocalFileSystem."""

    def expand_path(
        self,
        path: Union[str, list[str]],
        recursive: bool = False,
        maxdepth: Optional[int] = None,
        disable_glob: bool = False,
        **kwargs: Any,
    ) -> Union[str, list[str]]:
        """Turn one or more globs or dirs into a list of files or dirs.

        kwargs are passed to ``glob`` or ``find``, which may in turn call ``ls``
        """
        # type: ignore
        if maxdepth is not None and maxdepth < 1:
            raise ValueError("maxdepth must be at least 1")

        if isinstance(path, (str, os.PathLike)):
            out = self.expand_path(
                [path], recursive, maxdepth, disable_glob=disable_glob
            )
        else:
            out = set()  # type: ignore
            path = [self._strip_protocol(p) for p in path]
            for p in path:
                if has_magic(p) and not disable_glob:
                    bit = set(self.glob(p, maxdepth=maxdepth, **kwargs))
                    out |= bit  # type: ignore
                    if recursive:
                        # glob call above expanded one depth so if maxdepth is defined
                        # then decrement it in expand_path call below. If it is zero
                        # after decrementing then avoid expand_path call.
                        if maxdepth is not None and maxdepth <= 1:
                            continue
                        out |= set(  # type: ignore
                            self.expand_path(
                                list(bit),
                                recursive=recursive,
                                maxdepth=maxdepth - 1 if maxdepth is not None else None,
                                **kwargs,
                            )
                        )
                    continue
                elif recursive:
                    rec = set(
                        self.find(
                            p, maxdepth=maxdepth, withdirs=True, detail=False, **kwargs
                        )
                    )
                    out |= rec  # type: ignore
                if p not in out and (recursive is False or self.exists(p)):
                    # should only check once, for the root
                    out.add(p)  # type: ignore
        if not out:
            raise FileNotFoundError(path)
        return sorted(out)


class FsspecCompatibleFS(AbstractFileSystem):
    """A fsspec compatible file system.

    Used to be compatible with old version in some special methods.
    """

    def walk(  # noqa
        self,
        path: str,
        maxdepth: Optional[int] = None,
        topdown: bool = True,
        on_error: str = "omit",
        **kwargs: Any,
    ) -> Any:
        """Return all files belows path.

        Copied from fsspec(2024.9.0) to fix fsspec(2023.5.0.)

        List all files, recursing into subdirectories; output is iterator-style,
        like ``os.walk()``. For a simple list of files, ``find()`` is available.

        When topdown is True, the caller can modify the dirnames list in-place (perhaps
        using del or slice assignment), and walk() will
        only recurse into the subdirectories whose names remain in dirnames;
        this can be used to prune the search, impose a specific order of visiting,
        or even to inform walk() about directories the caller creates or renames before
        it resumes walk() again.
        Modifying dirnames when topdown is False has no effect. (see os.walk)

        Note that the "files" outputted will include anything that is not
        a directory, such as links.

        Parameters
        ----------
        path: str
            Root to recurse into
        maxdepth: int
            Maximum recursion depth. None means limitless, but not recommended
            on link-based file-systems.
        topdown: bool (True)
            Whether to walk the directory tree from the top downwards or from
            the bottom upwards.
        on_error: "omit", "raise", a collable
            if omit (default), path with exception will simply be empty;
            If raise, an underlying exception will be raised;
            if callable, it will be called with a single OSError instance as argument
        kwargs: passed to ``ls``

        """
        # type: ignore
        if maxdepth is not None and maxdepth < 1:
            raise ValueError("maxdepth must be at least 1")

        path = self._strip_protocol(path)
        full_dirs = {}
        dirs = {}
        files = {}

        detail = kwargs.pop("detail", False)
        try:
            listing = self.ls(path, detail=True, **kwargs)
        except (FileNotFoundError, OSError) as e:
            if on_error == "raise":
                raise
            elif callable(on_error):
                on_error(e)
            if detail:
                return path, {}, {}  # type: ignore
            return path, [], []  # type: ignore

        for info in listing:
            # each info name must be at least [path]/part , but here
            # we check also for names like [path]/part/
            pathname = info["name"].rstrip("/")  # type: ignore
            name = pathname.rsplit("/", 1)[-1]
            if info["type"] == "directory" and pathname != path:  # type: ignore
                # do not include "self" path
                full_dirs[name] = pathname
                dirs[name] = info
            elif pathname == path:
                # file-like with same name as give path
                files[""] = info
            else:
                files[name] = info

        if not detail:
            dirs = list(dirs)  # type: ignore
            files = list(files)  # type: ignore

        if topdown:
            # Yield before recursion if walking top down
            yield path, dirs, files

        if maxdepth is not None:
            maxdepth -= 1
            if maxdepth < 1:
                if not topdown:
                    yield path, dirs, files
                return

        with ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(
                    self.walk,
                    full_dirs[d],
                    maxdepth=maxdepth,
                    detail=detail,
                    topdown=topdown,
                    **kwargs,
                ): d
                for d in dirs
            }
            for future in as_completed(futures):
                for result in future.result():
                    yield result

        if not topdown:
            # Yield after recursion if walking bottom up
            yield path, dirs, files

    def find(  # noqa #
        self,
        path: str,
        maxdepth: Optional[int] = None,
        withdirs: bool = False,
        detail: bool = False,
        **kwargs: Any,  # type: ignore
    ) -> Any:
        """List all files below path.

        Copied from fsspec(2024.9.0) to fix fsspec(2023.5.0.)

        Like posix ``find`` command without conditions

        Parameters
        ----------
        path : str
        maxdepth: int or None
            If not None, the maximum number of levels to descend
        withdirs: bool
            Whether to include directory paths in the output. This is True
            when used by glob, but users usually only want files.
        kwargs are passed to ``ls``.

        """
        # TODO: allow equivalent of -name parameter
        path = self._strip_protocol(path)
        out = {}

        # Add the root directory if withdirs is requested
        # This is needed for posix glob compliance
        if withdirs and path != "" and self.isdir(path):
            out[path] = self.info(path)

        for _, dirs, files in super().walk(path, maxdepth, detail=True, **kwargs):
            if withdirs:
                files.update(dirs)
            out.update({info["name"]: info for name, info in files.items()})
        if not out and self.isfile(path):
            # walk works on directories, but find should also return [path]
            # when path happens to be a file
            out[path] = {}
        names = sorted(out)
        if not detail:
            return names
        else:
            return {name: out[name] for name in names}

    def put(
        self,
        lpath: str,
        rpath: str,
        recursive: bool = False,
        callback: Any = None,
        maxdepth: Optional[int] = None,
        disable_glob: bool = False,
        **kwargs: Any,
    ) -> None:
        """Copy file(s) from local.

        Copies a specific file or tree of files (if recursive=True). If rpath
        ends with a "/", it will be assumed to be a directory, and target files
        will go within.

        Calls put_file for each source.
        """
        if isinstance(lpath, list) and isinstance(rpath, list):
            # No need to expand paths when both source and destination
            # are provided as lists
            rpaths = rpath
            lpaths = lpath
        else:
            from fsspec.implementations.local import make_path_posix, trailing_sep

            source_is_str = isinstance(lpath, str)
            if source_is_str:
                lpath = make_path_posix(lpath)
            fs = EnhancedLocalFileSystem()
            lpaths = fs.expand_path(
                lpath, recursive=recursive, maxdepth=maxdepth, disable_glob=disable_glob
            )
            if source_is_str and (not recursive or maxdepth is not None):
                # Non-recursive glob does not copy directories
                lpaths = [p for p in lpaths if not (trailing_sep(p) or fs.isdir(p))]
                if not lpaths:
                    return

            source_is_file = len(lpaths) == 1
            dest_is_dir = isinstance(rpath, str) and (
                trailing_sep(rpath) or self.isdir(rpath)
            )

            rpath = (
                self._strip_protocol(rpath)
                if isinstance(rpath, str)
                else [self._strip_protocol(p) for p in rpath]
            )
            exists = source_is_str and (
                (has_magic(lpath) and source_is_file)
                or (
                    (not has_magic(lpath) or disable_glob)
                    and dest_is_dir
                    and not trailing_sep(lpath)
                )
            )
            rpaths = other_paths(
                lpaths,
                rpath,
                exists=exists,
                flatten=not source_is_str,
            )

        for lpath, rpath in zip(lpaths, rpaths):
            self.put_file(lpath, rpath, **kwargs)
